from fs.enums import ResourceType
from fs.errors import ResourceReadOnly, ResourceNotFound
from fs.info import Info

from .baseresource import BaseAnVILResource
from .google import GoogleAnVILFile, DRSAnVILFile


class BaseAnVILFolder(BaseAnVILResource):
    def __init__(self, name, last_modified=None):
        self.initialized = False
        # required as anvil supports files that have same name as directory
        if name[-1] != "/":
            self.name = name + "/"
        else:
            self.name = name
        self.last_modified = last_modified
        self.children = {}

    def lazily_init(fn):
        def lazywrapper(*args, **kwargs):
            self = args[0]
            if not self.initialized:
                self.lazy_init()
                self.initialized = True
            return fn(*args, **kwargs)
        return lazywrapper

    def lazy_init(self):
        raise NotImplementedError(
            f"{self.__class__.__name__}.lazy_init method is abstract")

    def __hash__(self):
        return hash((self.name, self.__class__.__name__, self.last_modified))

    def __eq__(self, other):
        return (self.name, self.last_modified) == (
            other.name, other.last_modified)

    # allow dictionary-style access, with possible objs as keys
    @lazily_init
    def __getitem__(self, key):
        return self.children[key]

    # allow for <item> in :
    def __iter__(self):
        return iter(self.children)

    @lazily_init
    def keys(self):
        def sorter(k):
            if k[-1] == "/":
                # space is first lexicographically
                k = "      "+k
            return k
        return sorted([k for k in self.children.keys()], key=sorter)

    @lazily_init
    def get_object_from_path(self, path):
        if path == "/" or path == "":
            return self
        # internally, using google-style no initial slash
        if path[0] == "/":
            path = path[1:]
        # if path represents a folder:
        if path[-1] == "/":
            split = path[:-1].split("/")
            for i in range(len(split)):
                split[i] += "/"
        else:
            split = path.split("/")
            for i in range(len(split[:-1])):
                split[i] += "/"
        base_obj = self
        for component in split:
            try:
                base_obj = base_obj[component]
            except KeyError:
                raise ResourceNotFound("Resource {} not found".format(path))
        return base_obj

    def __setitem__(self, key, val):
        self.children[key] = val

    def getinfo(self):
        _result = {
            "basic": {
                "name": self.name,
                "is_dir": True,
            },
            "details": {
                "type": ResourceType.directory,
                "modified": self.last_modified
            }
        }
        return Info(_result)

    def upload(self, fname, file):
        raise ResourceReadOnly(f"Error: {self.name} is read-only")

    def is_linkable_file(self, fname):

        protocol = fname.split("://")[0]
        filename = fname.split("/")[-1]
        allowed_protocols = {
            "gs": GoogleAnVILFile,
            "drs": DRSAnVILFile,
        }
        blocked_file_prefixes = [
            "data-explorer?"
        ]
        if protocol in allowed_protocols:
            for bfp in blocked_file_prefixes:
                if filename.startswith(bfp):
                    return None
            return allowed_protocols[protocol]
        else:
            return None



from .basefile import BaseAnVILFile
from .basefolder import BaseAnVILFolder
from .google import GoogleAnVILFile

import gs_chunked_io as gscio


class ReferenceDataFolder(BaseAnVILFolder):
    def __init__(self, name, refs):
        super().__init__(name)
        self.refs = refs

    def lazy_init(self):
        self.init_references(self.refs)

    def init_references(self, refs):
        for source in refs:
            # source, e.g. hg38
            source_folder = RefereneDataSubfolder(source+"/", refs[source])
            self[source_folder.name] = source_folder
            # reftype, e.g. axiomPoly_resource_vcf


class RefereneDataSubfolder(BaseAnVILFolder):
    def __init__(self, name, refs_source={}):
        super().__init__(name)
        self.refs_source = refs_source

    def lazy_init(self):
        for reftype in self.refs_source:
            reftype_folder = RefereneDataSubfolder(reftype+"/")
            self[reftype_folder.name] = reftype_folder
            contents = ReferenceDataFile.make_rdfs(
                self.refs_source[reftype])
            for c in contents:
                reftype_folder[c.name] = c


class ReferenceDataFile(BaseAnVILFile):
    def __init__(self, blob):
        blob.reload()
        self.blob_handle = blob
        self.name = blob.name.split("/")[-1]
        self.last_modified = blob.updated
        self.size = blob.size
        self.is_dir = False

    @classmethod
    def make_rdfs(cls, objs):
        return GoogleAnVILFile.factory(objs)

    def get_bytes_handler(self):
        return gscio.Reader(self.blob_handle)



from .basefolder import BaseAnVILFolder
from .workspace import Workspace


class Namespace(BaseAnVILFolder):
    def __init__(self, namespace_name, workspaces=[]):
        super().__init__(namespace_name)
        self.workspace_names = workspaces

    def lazy_init(self):
        for ws in self.workspace_names:
            self.fetch_workspace(ws)

    def fetch_workspace(self, workspace_name):
        ws = Workspace(self.name, workspace_name)
        self[ws.name] = ws
        return ws

    def __str__(self):
        out = "<Namespace {}>".format(self.name)
        # namespace can only have workspace folders, so keys are workspaces
        for ws in self.keys():
            out += "\n - {}".format(str(ws))
        return out



from anvilfs.anvilfs import Namespace


class TestNamespace:
    def test_init(self):
        name = "Space Name"
        ns = Namespace(name)
        assert ns.name == name + "/"

    def test_fetch_workspace(self, valid_gs_info):
        info = valid_gs_info
        ws = Namespace(info["Namespace"]).fetch_workspace(
            info["Workspace"]
        )
        assert ws.name == info["Workspace"] + "/"

    def test_str(self):
        name = "nsname"
        ns = Namespace(name)
        assert str(ns) == f"<Namespace {name}/>"



from anvilfs.anvilfs import Workspace


class TestWorkspace:
    def test_init(self, valid_gs_info):
        ns_name = valid_gs_info["Namespace"]
        ws_name = valid_gs_info["Workspace"]
        ws = Workspace(ns_name, ws_name)
        ws_d = ws.__dict__
        assert (
            ws.name == f"{ws_name}/" and
            "bucket_name" in ws_d and
            "attributes" in ws_d
        )

    def test_lazy_init(self, valid_gs_info):
        # things that will always be present:
        init_folders = [
            "Other Data/",
            "Reference Data/",
            "Tables/"
        ]
        ns_name = valid_gs_info["Namespace"]
        ws_name = valid_gs_info["Workspace"]
        ws = Workspace(ns_name, ws_name)
        assert (
            ws.keys() == init_folders
        )



import pytest
import requests

from anvilfs.baseresource import BaseAnVILResource, ClientRepository
from anvilfs.basefolder import BaseAnVILFolder
from anvilfs.basefile import BaseAnVILFile

from fs.enums import ResourceType
from fs.info import Info


class TestClientRepository:
    br = ClientRepository()

    def test_base_default_project(self):
        assert self.br.base_project is None

    def test_ref_defaults(self):
        rs = self.br._refs
        assert (
            rs["fapi"].__name__ == "firecloud.api" and
            rs["gc_storage_client"] is None and
            rs["gc_bigquery_client"] is None
        )

    def test_ref_inits(self):
        ris = self.br._ref_inits
        gcs_string = "<class 'google.cloud.storage.client.Client'>"
        gcbq_string = "<class 'google.cloud.bigquery.client.Client'>"
        assert (
            str(ris["gc_storage_client"]) == gcs_string and
            str(ris["gc_bigquery_client"]) == gcbq_string
        )

    def test_fapi_token(self):
        token = self.br.get_fapi_token()
        rs = ("https://www.googleapis.com/oauth2/v1/"
              f"tokeninfo?access_token={token}")
        r = requests.get(rs)
        assert r.status_code == 200


class TestBaseResource:

    bar = BaseAnVILResource()

    def test_info_abstract(self):
        with pytest.raises(NotImplementedError):
            self.bar.getinfo()

    def test_hash_abstract(self):
        with pytest.raises(KeyError):
            self.bar.__hash__()

    def test_eq_abstract(self):
        with pytest.raises(NotImplementedError):
            self.bar.__eq__(self.bar)

    def test_str_abstract(self):
        with pytest.raises(KeyError):
            self.bar.__str__()

    def test_ne_abstract(self):
        with pytest.raises(NotImplementedError):
            self.bar.__ne__(self.bar)


class TestBaseAnVILFolder:

    def test_init(self):
        name = "Steve the folder"
        last_modified = "Never"
        baf = BaseAnVILFolder(name, last_modified)
        assert (
            baf.name == name + "/" and
            baf.last_modified == last_modified and
            baf.children == {}
        )

    def test_lazy_init_abstract(self):
        baf = BaseAnVILFolder("Francesca Folder")
        with pytest.raises(NotImplementedError):
            baf.lazy_init()

    def test_hash(self):
        name = "Bort"
        last_modified = "Half past never"
        hashed = BaseAnVILFolder(name, last_modified).__hash__()
        assert hashed == hash((name + "/", "BaseAnVILFolder", last_modified))

    def test_eq(self):
        name1 = "Bort"
        lm1 = "Never"
        name2 = "Bort"
        lm2 = "Never"
        baf1 = BaseAnVILFolder(name1, lm1)
        baf2 = BaseAnVILFolder(name2, lm2)
        assert baf1 == baf2

    def test_ne(self):
        name1 = "Bort"
        lm1 = "Never"
        name2 = "Bart"
        lm2 = "Never"
        baf1 = BaseAnVILFolder(name1, lm1)
        baf2 = BaseAnVILFolder(name2, lm2)
        assert baf1 != baf2

    def test_keys(self):
        # requires initialized object
        baf = BaseAnVILFolder("Qwerty")
        baf.initialized = True  # circumvent abstract init
        baf["A"] = "1"
        baf["B"] = "2"
        assert baf.keys() == ["A", "B"]

    def test_setitem(self):
        baf = BaseAnVILFolder("Qwerty")
        baf["A"] = "1"
        assert baf.children["A"] == "1"

    def test_getitem(self):
        # requires initialized object
        baf = BaseAnVILFolder("Qwerty")
        baf.initialized = True  # circumvent abstract init
        baf.children["A"] = "1"
        assert baf["A"] == "1"

    def test_get_object_from_path(self):
        # requires initialized object
        name = "Qwerty"
        subname = "Asdf"
        baf = BaseAnVILFolder(name)
        baf.initialized = True  # circumvent abstract init
        subbaf = BaseAnVILFolder(subname)
        subbaf.initialized = True
        baf[subbaf.name] = subbaf
        subbaf["A"] = "1"
        assert baf.get_object_from_path(subname + "/A") == "1"

    def test_iter(self):
        baf = BaseAnVILFolder("Hello")
        for x in [("A", "1"), ("B", "2"), ("C", "3")]:
            baf[x[0]] = x[1]
        keys = [x for x in baf]
        assert keys == ["A", "B", "C"]

    def test_getinfo(self):
        name = "beep"
        lm = None
        baf = BaseAnVILFolder(name, lm)
        info = baf.getinfo()
        expected_info = Info({
            "basic": {
                "name": name + "/",
                "is_dir": True,
            },
            "details": {
                "type": ResourceType.directory,
                "modified": lm
            }
        })
        assert (
            info == expected_info
        )

    def test_is_linkable_file(self):
        gs = "gs://some-bucket/some/file"
        drs = "drs://some-bucket/some/file"
        nada = "hey i'm a string look at meeee"
        baf = BaseAnVILFolder("A")
        assert (
            baf.is_linkable_file(gs).__name__ == "GoogleAnVILFile" and
            baf.is_linkable_file(drs).__name__ == "DRSAnVILFile" and
            baf.is_linkable_file(nada) is None
        )


class TestBaseAnVILFile:
    def test_init(self):
        name = "bort"
        size = 2
        lm = "Last Blurnsday"
        baf = BaseAnVILFile(name, size, lm)
        assert (
            baf.name == name and
            baf.size == size and
            baf.last_modified == lm
        )

    def test_getinfo(self):
        name = "itchy"
        size = 3
        lm = "Boxing Day"
        baf = BaseAnVILFile(name, size, lm)
        i = Info(
            {
                "basic": {
                    "name": name,
                    "is_dir": False,
                },
                "details": {
                    "type": ResourceType.file,
                    "size": size,
                    "modified": lm
                }
            }
        )
        assert baf.getinfo() == i

    def test_string_to_buffer(self):
        string = "Strrriiiiinnggggg"
        baf = BaseAnVILFile("A", len(string))
        buff = baf.string_to_buffer(string)
        assert buff.read() == string.encode('utf-8')

    def test_get_bytes_handler(self):
        with pytest.raises(NotImplementedError):
            BaseAnVILFile("1", 2).get_bytes_handler()



try:
    from anvilfs.anvilfs import AnVILFS
except ImportError:
    AnVILFS = None
from ._pyfilesystem2 import PyFilesystem2FilesSource


class AnVILFilesSource(PyFilesystem2FilesSource):
    plugin_type = 'anvil'
    required_module = AnVILFS
    required_package = "fs.anvilfs"

    def _open_fs(self, user_context):
        props = self._serialization_props(user_context)
        handle = AnVILFS(**props)
        return handle


__all__ = ('AnVILFilesSource',)



from .basefolder import BaseAnVILFolder
from .reference import ReferenceDataFolder
from .tables import RootTablesFolder
from .workspacebucket import OtherDataFolder


class Workspace(BaseAnVILFolder):
    def __init__(self, namespace_name,  workspace_name):
        self.namespace_name = namespace_name
        # collect workspace attributes that inform structure
        resp = self.fetch_api_info(workspace_name)
        self.bucket_name = resp["workspace"]["bucketName"]
        self.attributes = resp["workspace"]["attributes"]
        try:
            super().__init__(
                workspace_name, resp["workspace"]["lastModified"])
        except KeyError:
            print("Error: Workspace fetch_api_info({}) fetch failed".format(
                workspace_name))

    def lazy_init(self):
        if self.initialized:
            print(f"{self.name} already initialized!")
            return
        # Tables folder
        table_baf = RootTablesFolder(self)
        self[table_baf.name] = table_baf
        # bucket folder
        bucket_baf = OtherDataFolder(self.attributes, self.bucket_name)
        self[bucket_baf.name] = bucket_baf
        # ref data folder
        refs = self.ref_extractor(self.attributes)
        ref_baf = ReferenceDataFolder("Reference Data/", refs)
        self[ref_baf.name] = ref_baf

    def ref_extractor(self, attribs):
        # structure:
        # { "source": {
        #      "reftype": [urlstr] }}
        result = {}

        for ref in [r for r in attribs if r.startswith("referenceData_")]:
            # e.g.,
            # referenceData_hg38_known_indels_sites_VCFs
            # source = hg38
            # reftype = known_indels_sites_VCFs
            val = attribs[ref]
            refsplit = ref.split("_", 2)
            source = refsplit[1]
            reftype = refsplit[2]

            # create entries if they dont exist...
            if source not in result:
                result[source] = {}
            if reftype not in result[source]:
                result[source][reftype] = []

            root_result_obj = result[source][reftype]

            # ensure its a list even if its a list of one
            if isinstance(val, dict):
                val = val["items"]
            elif isinstance(val, str):
                val = [val]

            # for every URI/URL...
            for v in val:
                allowed_protocols = [
                    "gs"
                ]
                if v.split("://")[0] in allowed_protocols:
                    root_result_obj.append(v)
        return result

    def fetch_api_info(self, workspace_name):
        fields = ("workspace.attributes,workspace.bucketName,"
                  "workspace.lastModified")
        resp = self.fapi.get_workspace(
            namespace=self.namespace_name, workspace=workspace_name,
            fields=fields)
        if resp.status_code == 200:
            return resp.json()
        else:
            resp.raise_for_status()

    def fetch_entity_info(self):
        resp = self.fapi.list_entity_types(
            namespace=self.namespace_name, workspace=self.name)
        if resp.status_code != 200:
            resp.raise_for_status()
        return resp.json()



from io import BytesIO
from os import SEEK_END, SEEK_SET

import re

import gs_chunked_io as gscio

from .basefile import BaseAnVILFile
from .basefolder import BaseAnVILFolder


class OtherDataFolder(BaseAnVILFolder):
    def __init__(self, attributes, bucket_name):
        super().__init__("Other Data")
        self.bucket_name = bucket_name
        self.attributes = attributes

    def lazy_init(self):
        # clone it to delete from one while iterating over other
        workspacedata = dict(self.attributes)
        blocklist_prefixes = [
            "referenceData_",
            "description",
            "tag:"
        ]
        for datum in self.attributes:
            for blocked in blocklist_prefixes:
                if datum.startswith(blocked):
                    del workspacedata[datum]
        if workspacedata:
            wsdf = WorkspaceDataFolder(workspacedata)
            self[wsdf.name] = wsdf
        _wsb = WorkspaceBucket(self.bucket_name)
        self[_wsb.name] = _wsb


class WorkspaceDataFolder(BaseAnVILFolder):
    def __init__(self, workspacedata):
        super().__init__("Workspace Data")
        self.workspacedata = workspacedata

    def lazy_init(self):
        files = {}
        for k in self.workspacedata:
            val = self.workspacedata[k]
            filetype = self.is_linkable_file(val)
            if filetype is not None:
                if filetype not in files:
                    files[filetype] = []
                files[filetype].append(val)
        linked_files = []
        for method in files:
            try:
                fresh_files = method.factory(files[method])
                linked_files.extend(fresh_files)
            except Exception as e:
                print("ERROR: SKIPPING FILE due to error:")
                print(e)
                continue
        linked_files.append(
            WorkspaceData("WorkspaceData.tsv", self.workspacedata))
        for f in linked_files:
            self[f.name] = f


class WorkspaceBucketSubFolder(BaseAnVILFolder):
    def __init__(self, name, bucketpath, bucket_name):
        self.bucket_name = bucket_name
        self.files = []
        self.bucketpath = bucketpath
        super().__init__(name)

    def lazy_init(self):
        pass

    def upload(self, fname, read_file):
        try:
            self["google_bucket"]
        except KeyError:
            self.google_bucket = self.gc_storage_client.bucket(
                self.bucket_name)
        with gscio.Writer(self.bucketpath + fname, self.google_bucket) as gsw:
            data = read_file.read(gsw.chunk_size)
            while data:
                gsw.write(data)
                data = read_file.read(gsw.chunk_size)


class WorkspaceBucket(BaseAnVILFolder):
    def __init__(self, bucket_name):
        super().__init__("Files")
        self.bucket_name = bucket_name
        self.bucketpath = ""

    def lazy_init(self):
        self.google_bucket = self.gc_storage_client.get_bucket(
            self.bucket_name)
        blobs = self.google_bucket.list_blobs()
        self.initialized = True
        for blob in blobs:
            self.insert_file(blob)

    def insert_file(self, bucket_blob):
        # name relative to the path from workspace bucket
        path = bucket_blob.name
        # handle subfolders like base folders -- dunno why google doesn't
        # e.g., list has a/b/ but not a/ 
        if path[-1] == "/":
            return
            #raise Exception(f"Files should be set, not folders: {path}")
        s = path.split("/")
        # march to terminal folder, creating along the way
        current = self
        for i in range(len(s)-1):
            subname = s[i]+'/'
            if subname not in current:
                dir_path = '/'.join(s[:i+1]) + '/'
                current[subname] = WorkspaceBucketSubFolder(subname, dir_path, self.bucket_name)
            current = current[subname]
        current[s[-1]] = WorkspaceBucketFile(bucket_blob)

    def upload(self, fname, read_file):
        try:
            self["google_bucket"]
        except KeyError:
            self.google_bucket = self.gc_storage_client.bucket(
                self.bucket_name)
        with gscio.Writer(fname, self.google_bucket) as gsw:
            data = read_file.read(gsw.chunk_size)
            while data:
                gsw.write(data)
                data = read_file.read(gsw.chunk_size)


class WorkspaceBucketFile(BaseAnVILFile):
    def __init__(self, blob):
        self.name = blob.name.split("/")[-1]
        self.size = blob.size
        self.last_modified = blob.updated
        self.blob_handle = blob
        self.is_dir = False

    def get_bytes_handler(self):
        return gscio.Reader(self.blob_handle)


class WorkspaceData(BaseAnVILFile):
    def __init__(self, name, data_dict):
        self.name = name
        self.buffer = self._dict_to_buffer(data_dict)
        self.last_modified = None

    def _dict_to_buffer(self, d):
        # only keys that match the below regex are valid
        keys = [k for k in d.keys() if bool(
            re.match("^[A-Za-z0-9_-]*$", k))]
        data = ""
        for k in keys:
            data += f"{k}\t{d[k]}\n"
        buffer = BytesIO(data.encode('utf-8'))
        position = buffer.tell()
        buffer.seek(0, SEEK_END)
        self.size = buffer.tell()
        buffer.seek(position, SEEK_SET)
        return buffer

    def get_bytes_handler(self):
        return self.buffer



import requests
import json
import concurrent.futures

from google.cloud import storage

import gs_chunked_io as gscio

from .basefile import BaseAnVILFile
from .clientrepository import ClientRepository

local_cr = ClientRepository()


class GoogleAnVILFile(BaseAnVILFile):
    def __init__(self, info):
        kb = 1024
        mb = 1024*kb
        self.chunk_size = 200*mb

        if type(info) == str and info.startswith("gs://"):
            self.blob = self.uri_to_blob(info)
            self.blob.reload()
            self.size = self.blob.size
            self.last_modified = self.blob.updated
            self.name = info.split("/")[-1]
        elif type(info) == dict:
            self.name = info["name"]
            if "blob" in info:
                self.blob = info["blob"]
            else:
                self.blob = self.info_to_blob(info["bucket"], info["path"])
            self.size = info["size"]
            self.last_modified = info["last_modified"]
        else:
            raise Exception(f"Bad GoogleAnVILFile init content:\n\t{info}")

    @classmethod
    def factory(cls, gs_uri_list):
        results = []
        MAX_BATCH_SIZE = 1000  # break into sub batches
        gs_uri_2d_array = []
        sublist = []
        LAZY_THRESHOLD = 1000  # if exceeded, fudge metadata for UX
        # create list of blob sublists, sublist <= 1000 elements
        if len(gs_uri_list) > LAZY_THRESHOLD:
            return [
                LazyGoogleAnVILFile(item) for item in gs_uri_list
            ]
        for gs_uri in gs_uri_list:
            # if max length reached, add it to 2d array
            sublist.append(cls.uri_to_blob(gs_uri))
            if len(sublist) == MAX_BATCH_SIZE:
                gs_uri_2d_array.append(sublist)
                sublist = []
        # add final under-sized batch to list if it exists
        if sublist:
            gs_uri_2d_array.append(sublist)
        # perform batch operations
        for batch in gs_uri_2d_array:
            with local_cr.gc_storage_client.batch():
                for item in batch:
                    item.reload()
            # sub list has been refreshed, create obj from metadata
            for item in batch:
                results.append(GoogleAnVILFile({
                    "name": item.name.split("/")[-1],
                    "last_modified": item.updated,
                    "size": item.size,
                    "blob": item
                }))
        return results

    @classmethod
    def uri_to_blob(cls, uri):
        split = uri.split("/")
        source_bucket = split[2]
        path = "/".join(split[3:])
        uproj = local_cr.gc_storage_client.project
        bucket = local_cr.gc_storage_client.bucket(
            source_bucket, user_project=uproj)
        return storage.blob.Blob(path, bucket)

    def info_to_blob(self, source_bucket, path):
        # requires project, bucket_name, prefix
        uproj = self.gc_storage_client.project
        bucket = self.gc_storage_client.bucket(
            source_bucket, user_project=uproj)
        return storage.blob.Blob(path, bucket)

    def get_bytes_handler(self):
        return gscio.Reader(self.blob)


# for use with very large lists
class LazyGoogleAnVILFile(GoogleAnVILFile):
    def __init__(self, uri, size=None, last_modified=None):
        self.uri = uri
        self.name = uri.split("/")[-1]
        if not size:
            self.size = 1
        else:
            self.size = size
        if not last_modified:
            self.last_modified = ""
        else:
            self.last_modified = last_modified

    def get_bytes_handler(self):
        super().__init__(self.uri)
        return super().get_bytes_handler()


class DRSAnVILFile(GoogleAnVILFile):
    api_url = ("https://us-central1-broad-dsde-prod.cloudfunctions.net/"
               "martha_v3")

    def __init__(self, input):
        token = ClientRepository().get_fapi_token()
        api_url = self.api_url
        response = requests.post(
            api_url,
            data={
                "url": input
            },
            headers={
                "Authorization": f"Bearer {token}"
            }
        )
        result = json.loads(response.text)
        gurl = result["gsUri"]
        super().__init__(gurl)

    @classmethod
    def factory(cls, drslist):
        # subfunction for threads
        def _load_url(drsuri, timeout):
            token = ClientRepository().get_fapi_token()
            url = cls.api_url
            r = requests.post(
                    url,
                    data={
                        "url": drsuri
                    },
                    headers={
                        "Authorization": f"Bearer {token}"
                    }
                )
            return json.loads(r.text)

        # thread pool maker
        def _pooler(inlist, maxworks=50):
            timeout = 60  # seconds
            good_data = []
            bad_uris = []
            with concurrent.futures.ThreadPoolExecutor(
                    max_workers=maxworks) as executor:
                # Start the load operations and mark each future with its URL
                future_to_url = {executor.submit(
                    _load_url, url, timeout): url for url in inlist}
                for future in concurrent.futures.as_completed(future_to_url):
                    url = future_to_url[future]
                    try:
                        data = future.result()
                        if "gsUri" not in data:
                            print(f"DRS resolution error- received:\n\t{data}")
                            bad_uris.append(url)
                        else:
                            good_data.append(data)
                    except Exception as exc:
                        print('%r generated an exception: %s' % (url, exc))
                        raise exc
                        bad_uris.append(url)
                    else:
                        pass
            return good_data, bad_uris

        # first pass
        file_objects = []
        good, bad = _pooler(drslist)
        # retry
        good_retries, bad_retries = _pooler(bad)
        if bad_retries:
            print(f"Unable to resolve the following URIs:\n{bad_retries}")
        total_goods = good + good_retries
        gs_info = [
            {
                "gsUri": x["gsUri"],
                "bucket": x["bucket"],
                "path": x["name"],
                "size": x["size"],
                "name": x["fileName"],
                "last_modified": x["timeUpdated"]
            }
            for x in total_goods]
        # make google bucket objects
        for item in gs_info:
            file_objects.append(GoogleAnVILFile(item))
        return file_objects


class LazyDRSAnVILFile(DRSAnVILFile):
    def __init__(self, uri, name, size=None, last_modified=None):
        self.uri = uri
        self.name = name
        if not size:
            self.size = 1
        else:
            self.size = size
        if not last_modified:
            self.last_modified = ""
        else:
            self.last_modified = last_modified

    def get_bytes_handler(self):
        super().__init__(self.uri)
        return super().get_bytes_handler()



import requests
from io import BytesIO

from .basefile import BaseAnVILFile


class HypertextAnVILFile(BaseAnVILFile):
    def __init__(self, url, name=None, size=None, last_modified=None):
        self.url = url
        if not name:
            self.name = url.split("/")[-1]
        if not size or not last_modified:
            r = requests.head(url)
            try:
                self.size = r.headers["Content-Length"]
            except KeyError:
                self.size = 1
            try:
                self.last_modified = r.headers["Last-Modified"]
            except KeyError:
                self.last_modified = ""

    def get_bytes_handler(self):
        return BytesIO(requests.get(self.url).content)

    @classmethod
    def factory(cls, urllist):
        results = []
        for item in urllist:
            results.append(HypertextAnVILFile(item))
        return results



from .clientrepository import ClientRepository
from .basefolder import BaseAnVILFolder
from .google import DRSAnVILFile
from .namespace import Namespace
from .workloadidentitycredentials import WorkloadIdentityCredentials
from .workspacebucket import WorkspaceBucket, WorkspaceBucketSubFolder

from google.auth.transport.requests import AuthorizedSession

from fs.base import FS
from fs.errors import DirectoryExpected, ResourceNotFound, FileExpected


class AnVILFS(FS, ClientRepository):
    DEFAULT_API_URL = "https://api.firecloud.org/api/"

    def __init__(self, namespace, workspace, api_url=None,
                 on_anvil=False, drs_url=None):
        super(AnVILFS, self).__init__()
        ClientRepository.base_project = namespace
        if not api_url:
            api_url = self.DEFAULT_API_URL
        else:
            self.fapi.fcconfig.set_root_url(api_url)
        # the API endpoint where DRS URI resolution requests are sent
        if drs_url:
            DRSAnVILFile.api_url = drs_url
        if on_anvil:
            scopes = ['https://www.googleapis.com/auth/userinfo.email',
                      'https://www.googleapis.com/auth/userinfo.profile',
                      'https://www.googleapis.com/auth/cloud-platform']
            credentials = WorkloadIdentityCredentials(scopes=scopes)
            self.fapi.__setattr__("__SESSION", AuthorizedSession(credentials))
        self.namespace = Namespace(namespace, [workspace])
        self.workspace = self.namespace[workspace+"/"]
        self.rootobj = self.workspace  # edit to make namespace root

    def getinfo(self, path, namespaces=None):
        return self.rootobj.get_object_from_path(path).getinfo()

    # Get a list of resource names (str) in a directory.
    def listdir(self, path):
        if path == "/" or path == "":
            return self.rootobj.keys()
        try:
            maybe_dir = self.rootobj.get_object_from_path(path)
        except KeyError:
            raise ResourceNotFound("Resource {} not found".format(path))
        if isinstance(maybe_dir, BaseAnVILFolder):
            return maybe_dir.keys()
        else:
            raise DirectoryExpected(f"{path}")

    def isdir(self, path):
        if path == "" or path[-1] != "/":
            path += "/"
        try:
            return self.getinfo(path).is_dir
        except ResourceNotFound:
            return False

    def scandir(self, path, **kwargs):
        if path[-1] != "/":
            path = path + "/"
        result = []
        dirlist = self.listdir(path)
        for o in dirlist:
            result.append(self.getinfo(path+o))
        return result

    def upload(self, path, file, **kwargs):
        cleaved = path.rsplit('/', 1)
        root_dir = cleaved[0] + "/"
        fname = cleaved[1]
        root_obj = self.rootobj.get_object_from_path(root_dir)
        root_obj.upload(fname, file)

    def makedir(self, path, **kwargs):  # Make a directory.
        raise Exception("makedir not implemented")

    def makedirs(self, path, **kwargs):  # Make directories.
        # google bucket 'directories' are just components of a filename,
        #   e.g. you cannot have an empty directory. so, making
        #   directories is entirely local to the plugin until file upload.
        #   here we just go through the objects to ensure that the last
        #   existent directory is a google bucket location and make the
        #   subdirs, but upload must 'create' the subdirs itself
        # strip leading and lagging slashes
        if path and path[0] == "/":
            path = path[1:]
        if path and path[-1] == "/":
            path = path[:-1]
        root_obj = self.rootobj
        dirs = path.split("/")
        if "" in dirs:
            raise Exception("Empty directory name in path " + 
                f"{path} not supported")
        for d in dirs:
            try:
                root_obj = root_obj[d + "/"]
            except KeyError:
                t = type(root_obj)
                if  t != WorkspaceBucket and t != WorkspaceBucketSubFolder:
                    raise Exception("Only workspace bucket files are writable")
                prev_bucketpath = root_obj.bucketpath
                bucketname = root_obj.bucket_name
                new_dir = WorkspaceBucketSubFolder(d, prev_bucketpath + d + "/",
                    bucketname)
                root_obj[d + "/"] = new_dir
                root_obj = root_obj[d + "/"]
        return
                
        

    def openbin(self, path, mode="r", buffering=-1, **options):
        obj = self.rootobj.get_object_from_path(path)
        try:
            return obj.get_bytes_handler()
        except AttributeError:
            raise FileExpected(
                "Error: requested object is not a file:\n  {}".format(path))

    def remove():  # Remove a file.
        raise Exception("remove not implemented")

    def removedir():  # Remove a directory.
        raise Exception("removedir not implemented")

    def setinfo():  # Set resource information.
        raise Exception("setinfo not implemented")



import firecloud.api as fapi
from google.cloud import storage
from google.cloud import bigquery


class ClientRepository:
    base_project = None

    _refs = {
        "fapi": fapi,
        "gc_storage_client": None,
        "gc_bigquery_client": None
    }
    _ref_inits = {
        "gc_storage_client": storage.Client,
        "gc_bigquery_client": bigquery.Client
    }

    def __getattr__(self, ref):
        if not self._refs[ref]:
            self._refs[ref] = self._ref_inits[ref](project=self.base_project)
        return self._refs[ref]

    def get_fapi_token(self):
        try:
            sesh = self.fapi.__getattribute__("__SESSION")
        except AttributeError:
            self.fapi._set_session()
        if not sesh or not sesh.credentials.valid:
            self.fapi._set_session()
        return self.fapi.__getattribute__("__SESSION").credentials.token



import os.path

import pytest
import yaml


conf = {}

conf_file = "config.yml"

path = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
try:
    with open(path + "/" + conf_file) as f:
        conf = yaml.safe_load(f.read())
except FileNotFoundError as fnfe:
    print("Configuration file not found -- are you sure it exists?")
    raise fnfe
except yaml.scanner.ScannerError as se:
    print("Configuration file is not valid yaml")
    raise se


@pytest.fixture(scope="session")
def valid_gs_info():
    return conf["anvil_info"]



from .clientrepository import ClientRepository


class BaseAnVILResource(ClientRepository):
    def getinfo(self):
        raise NotImplementedError("Method getinfo() not implemented")

    def __hash__(self):
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other):
        return self.getinfo().raw == other.getinfo().raw

    def __str__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.name)

    def __ne__(self, other):
        return not(self == other)
